import inept


def test_simple():
    class Config(inept.Config):
        with _.group:
            a: int = 42

    c1 = Config()
    c2 = Config()
    c1['a'] = 6
    assert (c2|c1).to_dict() == {'a': 6}
    assert (c1|c2).to_dict() == {'a': 6}


def test_merge_setitem_win():

    class MenuPasCher(inept.Config):
        " Le menu le moins cher "
        with _.group:
            plat: str = "steak frite"
            with _.switch:
                entree: str
                dessert: str
            cafe: bool = False

    c1 = MenuPasCher()
    c2 = MenuPasCher()
    c1['dessert'] = "mousse au chocolat"
    c2['plat'] = "ravioli"
    assert c1.to_dict() == {'plat': 'steak frite', 'dessert': "mousse au chocolat", "cafe": False}
    assert c2.to_dict() == {'plat': 'ravioli', "cafe": False}
    assert (c1|c2).to_dict() == {'plat': 'ravioli', 'dessert': "mousse au chocolat", "cafe": False}
    assert (c2|c1).to_dict() == {'plat': 'ravioli', 'dessert': "mousse au chocolat", "cafe": False}
    assert c1.to_dict() == {'plat': 'steak frite', 'dessert': "mousse au chocolat", "cafe": False}
    assert c2.to_dict() == {'plat': 'ravioli', "cafe": False}
    c2['cafe'] = True
    assert c2.to_dict() == {'plat': 'ravioli', "cafe": True}
    assert (c1|c2).to_dict() == {'plat': 'ravioli', 'dessert': "mousse au chocolat", "cafe": True}
    assert (c2|c1).to_dict() == {'plat': 'ravioli', 'dessert': "mousse au chocolat", "cafe": True}
    assert c1.to_dict() == {'plat': 'steak frite', 'dessert': "mousse au chocolat", "cafe": False}
    assert c2.to_dict() == {'plat': 'ravioli', "cafe": True}
    c1['entree'] = 'soupe'
    assert c1.to_dict() == {'plat': 'steak frite', 'entree': "soupe", "cafe": False}
    assert (c1|c2).to_dict() == {'plat': 'ravioli', 'entree': "soupe", "cafe": True}
    assert (c2|c1).to_dict() == {'plat': 'ravioli', 'entree': "soupe", "cafe": True}
    assert c1.to_dict() == {'plat': 'steak frite', 'entree': "soupe", "cafe": False}
    assert c2.to_dict() == {'plat': 'ravioli', "cafe": True}




import pytest

import inept
from test_config_simple import test_conf_script, test_make_empty_conf


def test_conf_cli():
    conf = test_conf_script()
    conf.load_cli([
        "--a2e.b2gf.c4od", "1e-4",
        "--a1ob", "False",
        "--a2e.b2gf.d3gf.e1oi", "20"
    ])

    assert conf['a1ob'] == False
    assert conf['a2e.b2gf.d3gf.e1oi'] == 20
    assert conf['a2e.b2gf.d3gf'] == True
    assert conf['a2e.b2gf.c5oi'] == 100
    assert conf['a2e.b2gf.c4od'] == 1e-4
    assert conf['a3g.b3oi'] == 30


def test_conf_cli_alone():
    conf = test_make_empty_conf()

    conf.load_cli([
        '--a1ob', 'True',
        "--a2e.b1ef.c1of", 'True',
        '--a2e.b2gf.d4of', 'True',
        '--a2e.b2gf.d3gf.e1oi', '10',
        '--a2e.b2gf.d3gf', 'True',
    ])

    assert conf['a1ob'] == True
    assert conf['a2e.b2gf.d3gf.e1oi'] == 10
    assert conf['a2e.b2gf.d3gf'] == True
    assert conf['a2e.b2gf.c5oi'] == 100
    assert conf['a3g.b3oi'] == 30


def test_help(capsys):
    conf = test_make_empty_conf(command_name='cli')

    with pytest.raises(SystemExit):
        conf.load_cli(["--help"])
    captured = capsys.readouterr()

    assert captured.out == """\
Usage: cli [OPTIONS]

Options:
  --a1ob BOOLEAN                a1ob doc
  --a2e.b1ef BOOLEAN            b1ef doc
  --a2e.b1ef.c1of BOOLEAN
  --a2e.b1ef.c2of BOOLEAN
  --a2e.b1ef.c3gf BOOLEAN
  --a2e.b1ef.c3gf.d1o TEXT      d1o doc
  --a2e.b1ef.c3gf.d2o TEXT
  --a2e.b2gf BOOLEAN
  --a2e.b2gf.d3gf BOOLEAN
  --a2e.b2gf.d3gf.e1oi INTEGER
  --a2e.b2gf.d4of BOOLEAN
  --a2e.b2gf.c4od FLOAT
  --a2e.b2gf.c5oi INTEGER
  --a2e.b2gf.c6o TEXT
  --a3g.b3oi INTEGER
  --a3g.b4od FLOAT
  --help                        Show this message and exit.
"""


def test_help_exposes_docstrings(capsys):

    class Config(inept.Config):
        "the class doc"
        with _.group:
            _ = "the main help"
            opt: str = ''
            _ = "an option"
            with _.switch as a:
                _ = "a switch"  # TODO: not included
                x: str = 'hello'
                _ = "the 'x' value"
                with _.options as y:
                    _ = "the 'y' value"
                    p: int = 0
                    _ = "the 'p' value"
                    q: int = 42
                    _ = "the 'q' value"

    conf = Config(command_name='cli')

    with pytest.raises(SystemExit):
        conf.load_cli(["--help"])
    captured = capsys.readouterr()

    assert captured.out == """\
Usage: cli [OPTIONS]

  the main help

Options:
  --opt TEXT       an option
  --a.x TEXT       the 'x' value
  --a.y BOOLEAN    the 'y' value
  --a.y.p INTEGER  the 'p' value
  --a.y.q INTEGER  the 'q' value
  --help           Show this message and exit.
"""




import sys
import os

PKG_RELATIVE_PATH = '..'


def add_package_to_path():
    conftest_path = os.path.dirname(__file__)
    pkg_path = os.path.abspath(os.path.join(conftest_path, PKG_RELATIVE_PATH))
    sys.path.insert(0, pkg_path)

add_package_to_path()



import collections


def flatten_path(nested, parent_key=()):
    items = []
    for k, v in nested.items():
        new_key = parent_key + (k,)
        if isinstance(v, collections.abc.MutableMapping):
            items.extend(flatten_path(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)


def flatten(nested, sep='.'):
    return {sep.join(k): v for k, v in flatten_path(nested).items()}



from inept import Config


def test_build():

    class MenuPasCher(Config):
        " Le menu le moins cher "
        with _.group:
            plat: str = "steak frite"
            with _.switch:
                entree: str
                dessert: str
            cafe: bool = False

    return MenuPasCher


def test_setitem():
    conf = test_build()()
    assert conf.to_dict() == {'plat': 'steak frite', 'cafe': False}
    conf['dessert'] = "mousse au chocolat"
    assert conf.to_dict() == {'plat': 'steak frite', 'cafe': False, 'dessert': 'mousse au chocolat'}
    conf['entree'] = "soupe"
    assert conf.to_dict() == {'plat': 'steak frite', 'cafe': False, 'entree': 'soupe'}


def test_carte():

    class Carte(Config):
        with _.options:
            cafe: bool = False
            boisson: str
            with _.switch:
                with _.group as menu_base:
                    plat: str = "steak frite"
                    with _.switch:
                        entree: str
                        dessert: str = "tarte"
                with _.group as menu_complet:
                    plat: str = "lasagnes"
                    entree: str = "salade"
                    dessert: str = "compote"

    c = Carte()

    assert list(c) == [
        'cafe',
        'boisson',
        'menu_base',
        'menu_base.plat',
        'menu_base.entree',
        'menu_base.dessert',
        'menu_complet',
        'menu_complet.plat',
        'menu_complet.entree',
        'menu_complet.dessert',
    ]

    assert c.to_dict() == {'cafe': False}

    c['boisson'] = "limonade"
    assert c.to_dict() == {'cafe': False, 'boisson': 'limonade'}

    c['menu_base'] = True
    assert c.to_dict() == {
        'cafe': False,
        'boisson': 'limonade',
        'menu_base': True,
        'menu_base.plat': 'steak frite',
        'menu_base.dessert': 'tarte',
    }

    c['menu_base.dessert'] = 'mousse au chocolat'
    assert c.to_dict() == {
        'cafe': False,
        'boisson': 'limonade',
        'menu_base': True,
        'menu_base.plat': 'steak frite',
        'menu_base.dessert': 'mousse au chocolat',
    }

    c['menu_base.entree'] = 'soupe'
    assert c.to_dict() == {
        'cafe': False,
        'boisson': 'limonade',
        'menu_base': True,
        'menu_base.plat': 'steak frite',
        'menu_base.entree': 'soupe',
    }

    c['menu_complet'] = True
    assert c.to_dict() == {
        'cafe': False,
        'boisson': 'limonade',
        'menu_complet': True,
        'menu_complet.entree': 'salade',
        'menu_complet.plat': 'lasagnes',
        'menu_complet.dessert': 'compote',
    }

    c['menu_complet'] = False
    assert c.to_dict() == {
        'cafe': False,
        'boisson': 'limonade',
    }

    c['menu_base'] = True
    assert c.to_dict() == {
        'cafe': False,
        'boisson': 'limonade',
        'menu_base': True,
        'menu_base.plat': 'steak frite',
        'menu_base.entree': 'soupe',
    }




import collections
from .types import Identity


NoDefault = object()


class Node:

    def __init__(self, name, type=bool, default=NoDefault, doc=None, nodes=()):
        self.name = None if name is None else str(name)
        self.doc = None if doc is None else str(doc)
        self.default = default
        self.nodes = list(nodes)
        self.type = type

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.name!r}>"

    @property
    def has_default(self):
        return self.default is not NoDefault

    def convert(self, value):
        type = self.type
        try:
            value = type(value)
        except Exception as err:
            raise ValueError(
                f"Wrong value ({value!r}) for {self!r}, "
                f"should be of type {type.__qualname__}."
            )
        return value

    def iter_edges(self):
        for child in self.nodes:
            yield (self, child)
            yield from child.iter_edges()

    def walk(self, filter=True):
        yield (self,)

    @staticmethod
    def name_from_path(path):
        assert all(isinstance(e, Node) for e in path)
        return '.'.join(e.name for e in path if e.name)

    def path_from_name(self, name):
        mapping = {Node.name_from_path(p): p for p in self.walk()}
        return mapping[name]

    def validate(self):
        pass

    def rename(self, name):
        res = object.__new__(type(self))
        res.__dict__.update(self.__dict__)
        res.name = None if name is None else str(name)
        return res


class Value(Node):

    def __init__(self, name, type=Identity, default=NoDefault, doc=None):
        assert name
        super().__init__(name, type=type, default=default, doc=doc)


class Group(Node):

    def __init__(self, name, nodes, default=NoDefault, doc=None):
        super().__init__(name, default=default, doc=doc, nodes=nodes)
        assert all(isinstance(e, Node) for e in nodes)

    def walk(self, filter=True):
        yield (self,)
        for node in self.nodes:
            for path in node.walk(filter):
                if filter:
                    head, *tail = path
                    if not tail and not isinstance(head, Value):
                        continue
                yield (self,) + path

    def validate(self):
        for child in self.nodes:
            if not isinstance(child, Node):
                raise ValueError(
                    f"nodes must {Node.__qualname__} objects, "
                    "get {child!r} instead."
                )
        names = [n.name for n in self.nodes]
        if len(names) != len(set(names)):
            dupes = [
                n for n, count in collections.Counter(names).items()
                if count > 1
            ]
            raise ValueError(f"duplicated names : {dupes}")
        for child in self.nodes:
            child.validate()


class Options(Group):

    def walk(self, filter=True):
        yield (self,)
        for node in self.nodes:
            for path in node.walk(filter):
                yield (self,) + path


class Switch(Options):

    def validate(self):
        super().validate()
        defaults = [
            n for n in self.nodes
            if n.has_default and not (isinstance(n, Group) and not n.default)
        ]
        if len(defaults) > 1:
            raise ValueError(
                "Can't have multiple nodes with default: "
                f"{list(defaults)}"
            )



import pytest
from inept import Value, Group, Options, Switch


def test_simple_Switch():
    root = Switch(None, [
        Group('a', [
            Value('x', int, 42),
            Value('y', str, 'hello'),
        ], True),
        Group('b', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ]),
    ])
    root.validate()
    assert list(map(str, root.walk())) == [
        "(<Switch None>,)",
        "(<Switch None>, <Group 'a'>)",
        "(<Switch None>, <Group 'a'>, <Value 'x'>)",
        "(<Switch None>, <Group 'a'>, <Value 'y'>)",
        "(<Switch None>, <Group 'b'>)",
        "(<Switch None>, <Group 'b'>, <Value 'x'>)",
        "(<Switch None>, <Group 'b'>, <Value 'y'>)",
    ]


def test_simple_Options():
    root = Options(None, [
        Group('a', [
            Value('x', int, 42),
            Value('y', str, 'hello'),
        ], True),
        Group('b', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ]),
    ])
    root.validate()
    assert list(map(str, root.walk())) == [
        "(<Options None>,)",
        "(<Options None>, <Group 'a'>)",
        "(<Options None>, <Group 'a'>, <Value 'x'>)",
        "(<Options None>, <Group 'a'>, <Value 'y'>)",
        "(<Options None>, <Group 'b'>)",
        "(<Options None>, <Group 'b'>, <Value 'x'>)",
        "(<Options None>, <Group 'b'>, <Value 'y'>)",
    ]


def test_simple_Group():
    root = Group(None, [
        Group('a', [
            Value('x', int, 42),
            Value('y', str, 'hello'),
        ], True),
        Group('b', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ]),
    ])
    root.validate()
    assert list(map(str, root.walk())) == [
        "(<Group None>,)",
        "(<Group None>, <Group 'a'>, <Value 'x'>)",
        "(<Group None>, <Group 'a'>, <Value 'y'>)",
        "(<Group None>, <Group 'b'>, <Value 'x'>)",
        "(<Group None>, <Group 'b'>, <Value 'y'>)",
    ]


def test_deeper_Switch():
    root = Switch(None, [
        Group('a', [
            Options('p', [
                Value('x', int, 42),
                Value('y', str, 'hello'),
            ]),
        ], True),
        Group('b', [
            Options('q', [
                Value('x', int, 6),
                Value('y', str, 'bye'),
            ]),
        ]),
    ])
    root.validate()
    assert list(map(str, root.walk())) == [
        "(<Switch None>,)",
        "(<Switch None>, <Group 'a'>)",
        "(<Switch None>, <Group 'a'>, <Options 'p'>, <Value 'x'>)",
        "(<Switch None>, <Group 'a'>, <Options 'p'>, <Value 'y'>)",
        "(<Switch None>, <Group 'b'>)",
        "(<Switch None>, <Group 'b'>, <Options 'q'>, <Value 'x'>)",
        "(<Switch None>, <Group 'b'>, <Options 'q'>, <Value 'y'>)",
    ]


def test_deeper_Options():
    root = Options(None, [
        Group('a', [
            Options('p', [
                Value('x', int, 42),
                Value('y', str, 'hello'),
            ]),
        ], True),
        Group('b', [
            Options('q', [
                Value('x', int, 6),
                Value('y', str, 'bye'),
            ]),
        ]),
    ])
    root.validate()
    assert list(map(str, root.walk())) == [
        "(<Options None>,)",
        "(<Options None>, <Group 'a'>)",
        "(<Options None>, <Group 'a'>, <Options 'p'>, <Value 'x'>)",
        "(<Options None>, <Group 'a'>, <Options 'p'>, <Value 'y'>)",
        "(<Options None>, <Group 'b'>)",
        "(<Options None>, <Group 'b'>, <Options 'q'>, <Value 'x'>)",
        "(<Options None>, <Group 'b'>, <Options 'q'>, <Value 'y'>)",
    ]


def test_deeper_Group():
    root = Group(None, [
        Group('a', [
            Options('p', [
                Value('x', int, 42),
                Value('y', str, 'hello'),
            ]),
        ], True),
        Group('b', [
            Options('q', [
                Value('x', int, 6),
                Value('y', str, 'bye'),
            ]),
        ]),
    ])
    root.validate()
    assert list(map(str, root.walk())) == [
        "(<Group None>,)",
        "(<Group None>, <Group 'a'>, <Options 'p'>, <Value 'x'>)",
        "(<Group None>, <Group 'a'>, <Options 'p'>, <Value 'y'>)",
        "(<Group None>, <Group 'b'>, <Options 'q'>, <Value 'x'>)",
        "(<Group None>, <Group 'b'>, <Options 'q'>, <Value 'y'>)",
    ]


def test_make_tree():
    root = Group(None, [
        Value("a1ob", bool, default=False, doc="a1ob doc"),
        Switch("a2e", [
            Switch('b1ef', doc="b1ef doc", nodes=[
                Value("c1of", bool),
                Value("c2of", bool),
                Group("c3gf", nodes=[
                    Value("d1o", doc="d1o doc"),
                    Value("d2o"),
                ]),
            ]),
            Group('b2gf', nodes=[
                Switch(None, [
                    Group("d3gf", nodes=[
                        Value("e1oi", int, default=50),
                    ]),
                    Value("d4of", bool),
                ]),
                Value("c4od", float, default=1e-8),
                Value("c5oi", int, default=100),
                Value("c6o"),
            ]),
        ]),
        Group("a3g", [
            Value("b3oi", int, default=30),
            Value("b4od", float, default=1e-6),
        ])
    ])
    root.validate()
    assert list(map(Value.name_from_path, root.walk())) == [
        '',
        'a1ob',
        'a2e.b1ef',
        'a2e.b1ef.c1of',
        'a2e.b1ef.c2of',
        'a2e.b1ef.c3gf',
        'a2e.b1ef.c3gf.d1o',
        'a2e.b1ef.c3gf.d2o',
        'a2e.b2gf',
        'a2e.b2gf.d3gf',
        'a2e.b2gf.d3gf.e1oi',
        'a2e.b2gf.d4of',
        'a2e.b2gf.c4od',
        'a2e.b2gf.c5oi',
        'a2e.b2gf.c6o',
        'a3g.b3oi',
        'a3g.b4od',
    ]
    return root


def test_invalid_duplicate():
    root = Switch(None, [
        Group('a', [
            Value('x', int, 42),
            Value('x', str, 'hello'),
        ], True),
        Group('b', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ]),
    ])
    with pytest.raises(ValueError):
        root.validate()


def test_invalid_duplicate_2():
    root = Switch(None, [
        Group('a', [
            Value('x', int, 42),
            Value('y', str, 'hello'),
        ], True),
        Group('a', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ]),
    ])
    with pytest.raises(ValueError):
        root.validate()


def test_invalid_Switch_multiple_activated():
    root = Switch(None, [
        Group('a', [
            Value('x', int, 42),
            Value('x', str, 'hello'),
        ], True),
        Group('b', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ], True),
    ])
    with pytest.raises(ValueError):
        root.validate()


def test_invalid_Switch_multiple_activated():
    root = Switch(None, [
        Group('a', [
            Value('x', int, 42),
            Value('y', str, 'hello'),
        ], True),
        Group('b', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ], False),
    ])
    root.validate()




class Identity:

    __click_type__ = str

    def __new__(cls, obj):
        return obj


# class Choice:
#     def __init__(self, values):
#         assert all(isinstance(x, Node) for x in values)
#         self.values = list(values)
# 
#     def __call__(self, value):
#         ...




from . import tree


class GroupContextManagerBase:

    def __init__(self, namespace, **kwds):
        self.namespace = namespace
        self.kwds = kwds

    def __enter__(self):
        self.namespace.enter(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.namespace.exit(self)

    @property
    def on(self):
        return type(self)(self.namespace, default=True)

    @property
    def off(self):
        return type(self)(self.namespace, default=False)

class GroupContextManager(GroupContextManagerBase):
    node_type = tree.Group


class OptionsContextManager(GroupContextManagerBase):
    node_type = tree.Options


class SwitchContextManager(GroupContextManagerBase):
    node_type = tree.Switch


class ContextManagerNamespace:
    def __init__(self, namespace):
        self.__namespace__ = namespace

    @property
    def group(self):
        return GroupContextManager(self.__namespace__)

    @property
    def options(self):
        return OptionsContextManager(self.__namespace__)

    @property
    def switch(self):
        return SwitchContextManager(self.__namespace__)


class AnnotationsDict(dict):

    def __init__(self, namespace):
        self.namespace = namespace

    def __setitem__(self, key, value):
        if self.namespace.level == 0:
            return super().__setitem__(key, value)
        self.namespace.add_type(key, value)


class TreeBuilderNamespace(dict):

    def __init__(self, magic_name):
        self._magic_name = str(magic_name)
        self.extra = {
            '__annotations__': AnnotationsDict(self),
            self._magic_name: ContextManagerNamespace(self),
        }
        self._stack = []
        self._entering = False
        self._groups = {}
        self._injected_nodes = set()
        self.root = None

    @property
    def level(self):
        return len(self._stack)

    @property
    def cm(self):
        return self._stack[-1] if self._stack else None

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            pass
        if key in self.extra:
            return self.extra[key]
        raise KeyError(key)

    def __setitem__(self, key, value):
        if self.level == 0:
            return super().__setitem__(key, value)
        if key == self._magic_name:
            return self.add_doc(value)
        entering, self._entering = self._entering, False
        if entering and value is self.cm:
            self._last_record.name = key
        else:
            self.record(key, value)

    def add_doc(self, value):
        assert isinstance(value, str)
        assert self._last_record.doc is None
        self._last_record.doc = value

    def add_type(self, key, value):
        node = self._last_record
        if node in self._injected_nodes:
            raise SyntaxError(
                "do not use annotation when assigning an existing tree"
                f" ('{key}: {value} = ...')"
            )
        if node.name == key:
            assert isinstance(node, tree.Value)
            node.type = value
        else:
            self.record(key, tree.NoDefault)
            self._last_record.type = value

    def parent(self, obj):
        if isinstance(obj, GroupContextManagerBase):
            if self.level >= 2:
                return self._stack[-2]
            else:
                return None
        return self.cm

    def record(self, name, obj):
        if isinstance(obj, GroupContextManagerBase):
            node = obj.node_type(name, [], **obj.kwds)
            self._groups[obj] = node
            if self.root is None:
                self.root = node
        elif isinstance(obj, tree.Node):
            node = obj.rename(name)
            self._injected_nodes.add(node)
        else:
            node = tree.Value(name, None, obj)
        parent = self.parent(obj)
        if parent:
            self._groups[parent].nodes.append(node)
        self._last_record = node

    def enter(self, cm):
        assert isinstance(cm, GroupContextManagerBase)
        self._stack.append(cm)
        self.record(None, cm)
        self._entering = True

    def exit(self, cm):
        assert cm is self._stack.pop()


class TreeBuilderMeta(type):

    __namespace_key__ = '_'

    @classmethod
    def __prepare__(mcls, name, bases, **kwds):
        return TreeBuilderNamespace(mcls.__namespace_key__)

    def __new__(mcls, name, bases, namespace):
        root = namespace.root
        if root is not None:
            root.validate()
        namespace = dict(namespace)
        namespace['root'] = root
        return type.__new__(mcls, name, bases, namespace)


class TreeBuilder(metaclass=TreeBuilderMeta):
    pass




from inept import ConfigSimple, Group, Switch, Options, Value


def test_default_under_flag_group():
    root = Options(None, [
        Group('a', nodes=[
            Value('x', int, 42),
            Value('y', bool, False)
        ]),
    ])

    conf = ConfigSimple(root)

    assert conf.to_dict() == {}
    assert conf.to_nested_dict() == {}

    conf['a'] = True

    assert conf.to_dict() == {
        'a': True,
        'a.x': 42,
        'a.y': False,
    }
    assert conf.to_nested_dict() == {
        'a': {
            'x': 42,
            'y': False,
        },
    }

    conf['a'] = False

    assert conf.to_dict() == {}
    assert conf.to_nested_dict() == {}

    conf['a.x'] = 2

    assert conf.to_dict() == {
        'a': True,
        'a.x': 2,
        'a.y': False,
    }
    assert conf.to_nested_dict() == {
        'a': {
            'x': 2,
            'y': False,
        },
    }




from inept.tree import Node
from inept.tree_builder import TreeBuilder


def test_build():

    class MyConfig(TreeBuilder):
        "MyConfig documentation, not for options"
        with _.group:
            _ = """
            general doc of option tree
            """
            a1ob: int = False
            _ = "a1ob doc"
            with _.switch as a2e:
                _ = "doc for a2e switch group"
                with _.switch as b1ef:
                    c1of: bool
                    c2of: bool
                    with _.group as c3gf:
                        d1o: str
                        d2o: str
                with _.group as b2gf:
                    with _.switch:
                        with _.group as d3gf:
                            e1oi: int = 50
                        d4of: bool
                    c4od: float = 1e-8
                    c5oi: int = 100
                    c6o: str
            with _.group as a3g:
                b3oi: int = 30
                b4od: float = 1e-6

    assert {Node.name_from_path(p) for p in MyConfig.root.walk()} == {
        '',
        'a1ob',
        'a2e.b1ef',
        'a2e.b1ef.c1of',
        'a2e.b1ef.c2of',
        'a2e.b1ef.c3gf',
        'a2e.b1ef.c3gf.d1o',
        'a2e.b1ef.c3gf.d2o',
        'a2e.b2gf',
        'a2e.b2gf.d3gf',
        'a2e.b2gf.d3gf.e1oi',
        'a2e.b2gf.d4of',
        'a2e.b2gf.c4od',
        'a2e.b2gf.c5oi',
        'a2e.b2gf.c6o',
        'a3g.b3oi',
        'a3g.b4od',
    }




import pytest
import inept


class MenuPasCher(inept.Config):
    with _.group:
        plat: str = "steak frite"
        with _.switch:
            entree: str
            dessert: str = "tarte"


class MenuComplet(inept.Config):
    with _.group:
        plat: str = "lasagnes"
        entree: str = "salade"
        dessert: str = "compote"


def test_simple():
    from inept import ConfigSimple, Value, Options, Switch
    carte_tree = Options(None, [
        Value('cafe', bool, False),
        Value('boisson', str),
        Switch(None, [
            MenuPasCher.root.rename("menu_base"),
            MenuComplet.root.rename("menu_complet"),
        ])
    ])
    order = ConfigSimple(carte_tree)
    assert list(order) == [
        'cafe',
        'boisson',
        'menu_base',
        'menu_base.plat',
        'menu_base.entree',
        'menu_base.dessert',
        'menu_complet',
        'menu_complet.plat',
        'menu_complet.entree',
        'menu_complet.dessert',
    ]


def test_tree_builder():

    class Carte(inept.Config):
        with _.options:
            cafe: bool = False
            boisson: str
            with _.switch:
                menu_base = MenuPasCher.root
                menu_complet = MenuComplet.root

    order = Carte()
    assert list(order) == [
        'cafe',
        'boisson',
        'menu_base',
        'menu_base.plat',
        'menu_base.entree',
        'menu_base.dessert',
        'menu_complet',
        'menu_complet.plat',
        'menu_complet.entree',
        'menu_complet.dessert',
    ]


def test_with_annotation():

    with pytest.raises(SyntaxError):

        class Carte(inept.Config):
            with _.options:
                cafe: bool = False
                boisson: str
                with _.switch:
                    menu_base: ... = MenuPasCher.root
                    menu_complet = MenuComplet.root




import sys
import collections
import click

from .tree import Node, Group, Switch, Options, Value


class ConfigMapping:

    root = Group(None, ())

    def __init__(self):
        self.validate_root(self.root)
        self._base_path = ()
        self._node_ids = {}
        # _data is (local) + (intial default)
        self._data = collections.ChainMap({}, self._init_default())

    def _register(self, node, id=None):
        if node in self._node_ids:
            if id is None:
                return
            raise ValueError(f"node {node} is already registered")
        assert isinstance(node, Node)
        if id is None:
            id = hash(node)
        self._node_ids[node] = id

    def _id(self, node):
        return self._node_ids.get(node, None)

    def _init_default(self):
        _id = self._id
        data = {}
        self._register(self.root)
        if isinstance(self.root, Group):
            data[_id(self.root)] = True
        elif isinstance(self.root, Value) and self.root.has_default:
            data[_id(self.root)] = self.root.convert(self.root.default)
        for parent, child in self.root.iter_edges():
            if child.has_default:
                value = child.default
            elif isinstance(child, Value):
                continue
            elif isinstance(parent, Options):
                value = False
            elif isinstance(parent, Group):
                value = True
            else:
                continue
            self._register(child)
            data[_id(child)] = child.convert(value)
        return data

    def validate_root(self, root):
        assert isinstance(root, Node)
        root.validate()

    def update(self, data):
        for key, value in dict(data).items():
            self[key] = value

    def __iter__(self):
        names = (Node.name_from_path(p) for p in self.root.walk())
        return (n for n in names if n)

    def info(self, msg):
        # FIXME On fait quoi pour les messages ? du logging ?
        print("INFO:", msg)

    def _get_path(self, key):
        return self._base_path + self.root.path_from_name(key)

    def __getitem__(self, key):
        if not key:
            raise KeyError(key)
        *path, target = self._get_path(key)
        path = map(self._id, path)
        id_target = self._id(target)
        if all(map(self._data.get, path)) and id_target in self._data:
            if isinstance(target, Value):
                return self._data[id_target]
            elif self._data[id_target]:
                return True
        raise KeyError(key)

    def __setitem__(self, key, value):
        if not key:
            raise KeyError(key)
        _id = self._id
        path = self._get_path(key)
        target = path[-1]
        for parent, child in zip(path[:-1], path[1:]):
            self._register(parent)
            self._data[_id(parent)] = True
            if isinstance(parent, Switch):
                actives = {n for n in parent.nodes if self._data.get(_id(n))}
                if actives - {child}:
                    olds = ' and '.join(repr(e.name) for e in actives)
                    self.info(f"{olds} is owerwritten by '{child.name}'")
                    for node in actives:
                        for dct in self._data.maps:
                            dct.pop(_id(node), None)
        self._register(target)
        self._data[_id(target)] = target.convert(value)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __delitem__(self, key):
        raise NotImplementedError("TODO")

    def __or__(self, other):
        assert self.root == other.root
        res = type(self)()
        res._node_ids = self._node_ids | other._node_ids
        res._data.maps[0].update(self._data.maps[0])
        res._data.maps[0].update(other._data.maps[0])
        return res


class ConfigExtract(ConfigMapping):

    def _get_base_path(self, node):
        for *base, leaf in self.root.walk(filter=False):
            if leaf == node:
                return tuple(base)
        raise ValueError(f"node {node} is not in self.root")

    def extract_node(self, node):
        base_path = self._get_base_path(node)
        root = node.rename(None)
        self._register(root, self._id(node))
        res = type(self)()
        res.root = root
        res._base_path = base_path
        res._node_ids = self._node_ids
        res._data = self._data
        return res


class ConfigCLI(ConfigMapping):

    def __init__(self, command_name=None, **kwds):
        super().__init__(**kwds)
        self.make_cli(command_name)

    @property
    def cli(self):
        return self._cli

    def make_cli(self, command_name):
        params = []
        for p in self.root.walk():
            name = Node.name_from_path(p)
            if not name:
                continue
            node = p[-1]
            opt = click.Option(
                [f"--{name}"],
                expose_value=False,
                type=self.get_type(node),
                help=node.doc,
            )
            opt.name = name
            opt.expose_value = True
            params.append(opt)
        help = self.root.doc
        self._cli = click.Command(command_name, params=params, help=help)

    @staticmethod
    def get_type(node):
        type = getattr(node, "type", None)
        type = getattr(type, '__click_type__', type)
        return type

    def parse_cli(self, args=None, **extra):
        if args is None:
            args = sys.argv[1:]
        ctx = self.cli.make_context(self.cli.name, args, **extra)
        return {k: v for k, v in ctx.params.items() if v is not None}

    def load_cli(self, args=None, **extra):
        try:
            self.update(self.parse_cli(args, **extra))
        except click.exceptions.Exit as err:
            sys.exit(err.exit_code)


class ConfigSerialize(ConfigMapping):

    def to_dict(self):
        res = {}
        for path in self.root.walk():
            key = Node.name_from_path(path)
            try:
                value = self[key]
            except KeyError:
                continue
            res[key] = value
        return res

    def to_nested_dict(self):
        res = {}
        for path in self.root.walk():
            key = Node.name_from_path(path)
            try:
                value = self[key]
            except KeyError:
                continue
            data = res
            for node in path[:-1]:
                if node.name:
                    data = data.setdefault(node.name, {})
            if isinstance(path[-1], Value):
                data[path[-1].name] = value
        return res


class ConfigBase(ConfigSerialize, ConfigCLI, ConfigExtract, ConfigMapping):
    pass


class ConfigSimple(ConfigBase):

    def __init__(self, root, **kwds):
        self._root = root
        super().__init__(**kwds)

    @property
    def root(self):
        return self._root





try:
    from ._version import version as __version__
except ModuleNotFoundError:
    pass

from .tree import Value, Group, Options, Switch
from .tree_builder import TreeBuilder
from .config import ConfigSimple, ConfigBase
from .utils import flatten


class Config(TreeBuilder, ConfigBase):
    pass




# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
import subprocess

import sphinx_rtd_theme


# -- Path setup --------------------------------------------------------------

ROOT = '../../'
sys.path.insert(0, os.path.abspath(ROOT))


# -- Project information -----------------------------------------------------

project = 'inept'
copyright = '2021, inept development team'
author = 'inept development team'

# The full version, including alpha/beta/rc tags
try:
    subprocess.check_call([sys.executable, "-m", "pip", "install", "-e", ROOT])
    from inept import __version__ as release
except:
    release = None


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'nbsphinx',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']
html_static_path = []


# -- setup() -----------------------------------------------------------------

def run_apidoc(_):
    from sphinx.ext.apidoc import main
    import os
    import shutil
    cur_dir = os.path.dirname(__file__)
    module = os.path.join(cur_dir, '../../inept')
    output_path = os.path.join(cur_dir, 'api')
    shutil.rmtree(output_path, ignore_errors=True)
    main(['--separate',
        '--module-first',
        '--no-toc',
        '--force',
        '-o', output_path, module,
    ])


def setup(app):
    app.connect('builder-inited', run_apidoc)




from setuptools import setup, find_packages


setup(
    name='inept',
    use_scm_version={'write_to': 'inept/_version.py'},
    setup_requires=['setuptools_scm', 'wheel'],
    description='INteractive Editable oPTions',
    author='Farid Smai',
    author_email='f.smai@brgm.fr',
    url='https://github.com/brgm/inept',
    license='GNU GPLv3',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        'click',
    ],
    python_requires='>=3.6',
)



import pytest

from inept import ConfigSimple, Group, Switch, Value, Options

from test_tree import test_make_tree


def test_make_empty_conf(**kwds):
    root = test_make_tree()
    conf = ConfigSimple(root, **kwds)
    return conf


def test_conf_type():
    conf = test_make_empty_conf()

    with pytest.raises(ValueError):
        conf['a2e.b2gf.d3gf.e1oi'] = 'toto'


def test_conf_script():
    conf = test_make_empty_conf()

    conf['a1ob'] = True
    conf["a2e.b1ef.c1of"] = True
    conf['a2e.b2gf.d4of'] = True
    conf['a2e.b2gf.d3gf.e1oi'] = 10
    conf['a2e.b2gf.d3gf'] = True

    assert conf['a1ob'] == True
    assert conf['a2e.b2gf.d3gf.e1oi'] == 10
    assert conf['a2e.b2gf.d3gf'] == True
    assert conf['a2e.b2gf.c5oi'] == 100
    assert conf['a3g.b3oi'] == 30

    return conf


def test_default_under_exclusif():
    root = Switch(None, [
        Value('x', int, 42),
        Value('y', bool)
    ])

    conf = ConfigSimple(root)

    assert conf['x'] == 42
    with pytest.raises(KeyError):
        conf['y']

    conf['y'] = True

    assert conf['y'] == True
    with pytest.raises(KeyError):
        conf['x']


def test_multiple_default_under_exclusif():
    root = Switch(None, [
        Value('x', int, 42),
        Value('y', bool, True)
    ])

    with pytest.raises(ValueError):
        ConfigSimple(root)


def test_default_group_under_exclusif():
    root = Switch(None, [
        Group('a', default=True, nodes=[
            Value('x', int, 42),
            Value('y', bool, True)
        ]),
        Group('b', nodes=[
            Value('x', int, 42),
            Value('y', bool, True)
        ]),
    ])

    conf = ConfigSimple(root)

    assert conf['a'] == True
    assert conf['a.x'] == 42
    assert conf['a.y'] == True
    with pytest.raises(KeyError):
        conf['b']
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': True}

    conf['b'] = True

    assert conf['b'] == True
    assert conf['b.x'] == 42
    assert conf['b.y'] == True
    with pytest.raises(KeyError):
        conf['a']
    assert conf.to_dict() == {'b': True, 'b.x': 42, 'b.y': True}

    conf['b'] = False
    assert conf.to_dict() == {}


def test_one_option():
    root = Options(None, [
        Group('a', nodes=[
            Value('x', int, 42),
            Value('y', bool, False)
        ]),
    ])

    conf = ConfigSimple(root)

    with pytest.raises(KeyError):
        conf['a']
    with pytest.raises(KeyError):
        conf['a.x']
    with pytest.raises(KeyError):
        conf['a.y']
    assert conf.to_dict() == {}

    conf['a'] = True

    assert conf['a'] == True
    assert conf['a.x'] == 42
    assert conf['a.y'] == False
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': False}

    conf['a'] = False

    with pytest.raises(KeyError):
        conf['a']
    with pytest.raises(KeyError):
        conf['a.x']
    with pytest.raises(KeyError):
        conf['a.y']
    assert conf.to_dict() == {}


def test_options():
    root = Options(None, [
        Group('a', [
            Value('x', int, 42),
            Value('y', str, 'hello'),
        ]),
        Group('b', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ]),
    ])

    conf = ConfigSimple(root)
    assert conf.to_dict() == {}

    conf['a'] = True
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': 'hello'}

    conf['a'] = False
    assert conf.to_dict() == {}

    conf['b.x'] = 12
    assert conf.to_dict() == {'b': True, 'b.x': 12, 'b.y': 'bye'}

    conf['a'] = True
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': 'hello', 'b': True, 'b.x': 12, 'b.y': 'bye'}


def test_options_with_default():
    root = Options(None, [
        Group('a', [
            Value('x', int, 42),
            Value('y', str, 'hello'),
        ], True),
        Group('b', [
            Value('x', int, 6),
            Value('y', str, 'bye'),
        ]),
    ])

    conf = ConfigSimple(root)
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': 'hello'}

    conf['a'] = True
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': 'hello'}

    conf['a'] = False
    assert conf.to_dict() == {}

    conf['b.x'] = 12
    assert conf.to_dict() == {'b': True, 'b.x': 12, 'b.y': 'bye'}

    conf['a'] = True
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': 'hello', 'b': True, 'b.x': 12, 'b.y': 'bye'}


# def test_not_all_flag_under_exclusive_exclusive():
#     root = Switch(None, [
#         Group('a', nodes=[
#             Value('x', int, 42),
#         ]),
#         Group('b', nodes=[
#             Value('x', int, 12),
#         ]),
#     ])
#     with pytest.raises(ValueError):
#         ConfigSimple(root)


# def test_not_any_flag_under_exclusive_exclusive():
#     root = Switch(None, [
#         Group('a', nodes=[
#             Value('x', int, 42),
#         ]),
#         Group('b', nodes=[
#             Value('x', int, 12),
#         ]),
#     ])
#     with pytest.raises(ValueError):
#         ConfigSimple(root)


def test_empty_conf_exclusive():
    conf = ConfigSimple(
        Switch(None, [
            Group('a', nodes=[
                Value('x', int, 42),
            ]),
            Group('b', nodes=[
                Value('x', int, 12),
            ]),
        ])
    )
    assert conf.to_dict() == {}


def test_empty_conf_default_under_exclusive():
    conf = ConfigSimple(
        Switch(None, [
            Group('a', default=True, nodes=[
                Value('x', int, 42),
            ]),
            Group('b', nodes=[
                Value('x', int, 12),
            ]),
        ])
    )
    assert conf.to_dict() == {'a': True, 'a.x': 42}




import inept


def test_simple_options():

    class Config(inept.Config):
        with _.options:
            with _.group as a:
                x: int
                y: int
            with _.group as b:
                x: int
                y: int

    node_a = Config.root.nodes[0]
    assert node_a.name == 'a'

    conf = Config()
    subc = conf.extract_node(node_a)

    assert list(conf) == ['a', 'a.x', 'a.y', 'b', 'b.x', 'b.y']
    assert list(subc) == ['x', 'y']

    assert conf.to_dict() == {}
    assert subc.to_dict() == {}

    conf['a.x'] = 42
    assert conf.to_dict() == {'a': True, 'a.x': 42}
    assert subc.to_dict() == {'x': 42}

    subc['y'] = 6
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': 6}
    assert subc.to_dict() == {'x': 42, 'y': 6}

    conf['b.y'] = 12
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': 6, 'b':True, 'b.y': 12}
    assert subc.to_dict() == {'x': 42, 'y': 6}


def test_simple_group():

    class Config(inept.Config):
        with _.group:
            with _.group as a:
                x: int
                y: int
            with _.group as b:
                x: int
                y: int

    node_a = Config.root.nodes[0]
    assert node_a.name == 'a'

    conf = Config()
    subc = conf.extract_node(node_a)

    assert list(conf) == ['a.x', 'a.y', 'b.x', 'b.y']
    assert list(subc) == ['x', 'y']

    assert conf.to_dict() == {}
    assert subc.to_dict() == {}

    conf['a.x'] = 42
    assert conf.to_dict() == {'a.x': 42}
    assert subc.to_dict() == {'x': 42}

    subc['y'] = 6
    assert conf.to_dict() == {'a.x': 42, 'a.y': 6}
    assert subc.to_dict() == {'x': 42, 'y': 6}

    conf['b.y'] = 12
    assert conf.to_dict() == {'a.x': 42, 'a.y': 6, 'b.y': 12}
    assert subc.to_dict() == {'x': 42, 'y': 6}


def test_simple_switch():

    class Config(inept.Config):
        with _.switch:
            with _.group as a:
                x: int
                y: int
            with _.group as b:
                x: int
                y: int

    node_a = Config.root.nodes[0]
    assert node_a.name == 'a'

    conf = Config()
    subc = conf.extract_node(node_a)

    assert list(conf) == ['a', 'a.x', 'a.y', 'b', 'b.x', 'b.y']
    assert list(subc) == ['x', 'y']

    assert conf.to_dict() == {}
    assert subc.to_dict() == {}

    conf['a.x'] = 42
    assert conf.to_dict() == {'a': True, 'a.x': 42}
    assert subc.to_dict() == {'x': 42}

    subc['y'] = 6
    assert conf.to_dict() == {'a': True, 'a.x': 42, 'a.y': 6}
    assert subc.to_dict() == {'x': 42, 'y': 6}

    conf['b.y'] = 12
    assert conf.to_dict() == {'b':True, 'b.y': 12}
    assert subc.to_dict() == {}

    subc['x'] = 5
    assert conf.to_dict() == {'a': True, 'a.x': 5, 'a.y': 6}
    assert subc.to_dict() == {'x': 5, 'y': 6}


def test_deep_options():

    class Config(inept.Config):
        with _.group:
            with _.options as a:
                with _.group as ap:
                    x: int
                    y: int
                with _.group as aq:
                    x: int
                    y: int
            with _.options as b:
                with _.group as bp:
                    x: int
                    y: int
                with _.group as bq:
                    x: int
                    y: int

    node_a = Config.root.nodes[0]
    node_ap = Config.root.nodes[0].nodes[0]
    assert node_ap.name == 'ap'

    conf = Config()
    c_a = conf.extract_node(node_a)
    c_ap = conf.extract_node(node_ap)

    assert list(conf) == [
        'a.ap', 'a.ap.x', 'a.ap.y',
        'a.aq', 'a.aq.x', 'a.aq.y',
        'b.bp', 'b.bp.x', 'b.bp.y',
        'b.bq', 'b.bq.x', 'b.bq.y',
    ]
    assert list(c_a) == [
        'ap', 'ap.x', 'ap.y',
        'aq', 'aq.x', 'aq.y',
    ]
    assert list(c_ap) == ['x', 'y']

    assert conf.to_dict() == {}
    assert c_a.to_dict() == {}
    assert c_ap.to_dict() == {}

    conf['a.ap.x'] = 42
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 42}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 42}
    assert c_ap.to_dict() == {'x': 42}

    c_a['ap.x'] = 21
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 21}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21}
    assert c_ap.to_dict() == {'x': 21}

    c_ap['y'] = 6
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 21, 'a.ap.y': 6}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}

    conf['b.bq.y'] = 12
    assert conf.to_dict() == {
        'a.ap': True, 'a.ap.x': 21, 'a.ap.y': 6, 'b.bq': True, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}


def test_deep_group():

    class Config(inept.Config):
        with _.group:
            with _.group as a:
                with _.group as ap:
                    x: int
                    y: int
                with _.group as aq:
                    x: int
                    y: int
            with _.group as b:
                with _.group as bp:
                    x: int
                    y: int
                with _.group as bq:
                    x: int
                    y: int

    node_a = Config.root.nodes[0]
    node_ap = Config.root.nodes[0].nodes[0]
    assert node_ap.name == 'ap'

    conf = Config()
    c_a = conf.extract_node(node_a)
    c_ap = conf.extract_node(node_ap)

    assert list(conf) == [
        'a.ap.x', 'a.ap.y',
        'a.aq.x', 'a.aq.y',
        'b.bp.x', 'b.bp.y',
        'b.bq.x', 'b.bq.y',
    ]
    assert list(c_a) == [
        'ap.x', 'ap.y',
        'aq.x', 'aq.y',
    ]
    assert list(c_ap) == ['x', 'y']

    assert conf.to_dict() == {}
    assert c_a.to_dict() == {}
    assert c_ap.to_dict() == {}

    conf['a.ap.x'] = 42
    assert conf.to_dict() == {'a.ap.x': 42}
    assert c_a.to_dict() == {'ap.x': 42}
    assert c_ap.to_dict() == {'x': 42}

    c_a['ap.x'] = 21
    assert conf.to_dict() == {'a.ap.x': 21}
    assert c_a.to_dict() == {'ap.x': 21}
    assert c_ap.to_dict() == {'x': 21}

    c_ap['y'] = 6
    assert conf.to_dict() == {'a.ap.x': 21, 'a.ap.y': 6}
    assert c_a.to_dict() == {'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}

    conf['b.bq.y'] = 12
    assert conf.to_dict() == {
        'a.ap.x': 21, 'a.ap.y': 6, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}


def test_deep_switch():

    class Config(inept.Config):
        with _.group:
            with _.switch as a:
                with _.group as ap:
                    x: int
                    y: int
                with _.group as aq:
                    x: int
                    y: int
            with _.switch as b:
                with _.group as bp:
                    x: int
                    y: int
                with _.group as bq:
                    x: int
                    y: int

    node_a = Config.root.nodes[0]
    node_ap = Config.root.nodes[0].nodes[0]
    assert node_ap.name == 'ap'

    conf = Config()
    c_a = conf.extract_node(node_a)
    c_ap = conf.extract_node(node_ap)

    assert list(conf) == [
        'a.ap', 'a.ap.x', 'a.ap.y',
        'a.aq', 'a.aq.x', 'a.aq.y',
        'b.bp', 'b.bp.x', 'b.bp.y',
        'b.bq', 'b.bq.x', 'b.bq.y',
    ]
    assert list(c_a) == [
        'ap', 'ap.x', 'ap.y',
        'aq', 'aq.x', 'aq.y',
    ]
    assert list(c_ap) == ['x', 'y']

    assert conf.to_dict() == {}
    assert c_a.to_dict() == {}
    assert c_ap.to_dict() == {}

    conf['a.ap.x'] = 42
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 42}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 42}
    assert c_ap.to_dict() == {'x': 42}

    c_a['ap.x'] = 21
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 21}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21}
    assert c_ap.to_dict() == {'x': 21}

    c_ap['y'] = 6
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 21, 'a.ap.y': 6}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}

    conf['b.bq.y'] = 12
    assert conf.to_dict() == {
        'a.ap': True, 'a.ap.x': 21, 'a.ap.y': 6, 'b.bq': True, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}

    c_a['aq.y'] = 7
    assert conf.to_dict() == {
        'a.aq': True, 'a.aq.y': 7, 'b.bq': True, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'aq': True, 'aq.y': 7}
    assert c_ap.to_dict() == {}

    c_ap['x'] = 5
    assert conf.to_dict() == {
        'a.ap': True, 'a.ap.x': 5, 'a.ap.y': 6, 'b.bq': True, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'ap': True, 'ap.x': 5, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 5, 'y': 6}


def test_deep_from_sub_options():

    class Config(inept.Config):
        with _.group:
            with _.options as a:
                with _.group as ap:
                    x: int
                    y: int
                with _.group as aq:
                    x: int
                    y: int
            with _.options as b:
                with _.group as bp:
                    x: int
                    y: int
                with _.group as bq:
                    x: int
                    y: int

    node_a = Config.root.nodes[0]
    node_ap = Config.root.nodes[0].nodes[0]
    assert node_ap.name == 'ap'

    conf = Config()
    c_a = conf.extract_node(node_a)
    c_ap = c_a.extract_node(node_ap)

    assert list(conf) == [
        'a.ap', 'a.ap.x', 'a.ap.y',
        'a.aq', 'a.aq.x', 'a.aq.y',
        'b.bp', 'b.bp.x', 'b.bp.y',
        'b.bq', 'b.bq.x', 'b.bq.y',
    ]
    assert list(c_a) == [
        'ap', 'ap.x', 'ap.y',
        'aq', 'aq.x', 'aq.y',
    ]
    assert list(c_ap) == ['x', 'y']

    assert conf.to_dict() == {}
    assert c_a.to_dict() == {}
    assert c_ap.to_dict() == {}

    conf['a.ap.x'] = 42
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 42}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 42}
    assert c_ap.to_dict() == {'x': 42}

    c_a['ap.x'] = 21
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 21}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21}
    assert c_ap.to_dict() == {'x': 21}

    c_ap['y'] = 6
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 21, 'a.ap.y': 6}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}

    conf['b.bq.y'] = 12
    assert conf.to_dict() == {
        'a.ap': True, 'a.ap.x': 21, 'a.ap.y': 6, 'b.bq': True, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}


def test_deep_from_sub_group():

    class Config(inept.Config):
        with _.group:
            with _.group as a:
                with _.group as ap:
                    x: int
                    y: int
                with _.group as aq:
                    x: int
                    y: int
            with _.group as b:
                with _.group as bp:
                    x: int
                    y: int
                with _.group as bq:
                    x: int
                    y: int

    node_a = Config.root.nodes[0]
    node_ap = Config.root.nodes[0].nodes[0]
    assert node_ap.name == 'ap'

    conf = Config()
    c_a = conf.extract_node(node_a)
    c_ap = c_a.extract_node(node_ap)

    assert list(conf) == [
        'a.ap.x', 'a.ap.y',
        'a.aq.x', 'a.aq.y',
        'b.bp.x', 'b.bp.y',
        'b.bq.x', 'b.bq.y',
    ]
    assert list(c_a) == [
        'ap.x', 'ap.y',
        'aq.x', 'aq.y',
    ]
    assert list(c_ap) == ['x', 'y']

    assert conf.to_dict() == {}
    assert c_a.to_dict() == {}
    assert c_ap.to_dict() == {}

    conf['a.ap.x'] = 42
    assert conf.to_dict() == {'a.ap.x': 42}
    assert c_a.to_dict() == {'ap.x': 42}
    assert c_ap.to_dict() == {'x': 42}

    c_a['ap.x'] = 21
    assert conf.to_dict() == {'a.ap.x': 21}
    assert c_a.to_dict() == {'ap.x': 21}
    assert c_ap.to_dict() == {'x': 21}

    c_ap['y'] = 6
    assert conf.to_dict() == {'a.ap.x': 21, 'a.ap.y': 6}
    assert c_a.to_dict() == {'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}

    conf['b.bq.y'] = 12
    assert conf.to_dict() == {
        'a.ap.x': 21, 'a.ap.y': 6, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}


def test_deep_from_sub_switch():

    class Config(inept.Config):
        with _.group:
            with _.switch as a:
                with _.group as ap:
                    x: int
                    y: int
                with _.group as aq:
                    x: int
                    y: int
            with _.switch as b:
                with _.group as bp:
                    x: int
                    y: int
                with _.group as bq:
                    x: int
                    y: int

    node_a = Config.root.nodes[0]
    node_ap = Config.root.nodes[0].nodes[0]
    assert node_ap.name == 'ap'

    conf = Config()
    c_a = conf.extract_node(node_a)
    c_ap = c_a.extract_node(node_ap)

    assert list(conf) == [
        'a.ap', 'a.ap.x', 'a.ap.y',
        'a.aq', 'a.aq.x', 'a.aq.y',
        'b.bp', 'b.bp.x', 'b.bp.y',
        'b.bq', 'b.bq.x', 'b.bq.y',
    ]
    assert list(c_a) == [
        'ap', 'ap.x', 'ap.y',
        'aq', 'aq.x', 'aq.y',
    ]
    assert list(c_ap) == ['x', 'y']

    assert conf.to_dict() == {}
    assert c_a.to_dict() == {}
    assert c_ap.to_dict() == {}

    conf['a.ap.x'] = 42
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 42}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 42}
    assert c_ap.to_dict() == {'x': 42}

    c_a['ap.x'] = 21
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 21}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21}
    assert c_ap.to_dict() == {'x': 21}

    c_ap['y'] = 6
    assert conf.to_dict() == {'a.ap': True, 'a.ap.x': 21, 'a.ap.y': 6}
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}

    conf['b.bq.y'] = 12
    assert conf.to_dict() == {
        'a.ap': True, 'a.ap.x': 21, 'a.ap.y': 6, 'b.bq': True, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'ap': True, 'ap.x': 21, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 21, 'y': 6}

    c_a['aq.y'] = 7
    assert conf.to_dict() == {
        'a.aq': True, 'a.aq.y': 7, 'b.bq': True, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'aq': True, 'aq.y': 7}
    assert c_ap.to_dict() == {}

    c_ap['x'] = 5
    assert conf.to_dict() == {
        'a.ap': True, 'a.ap.x': 5, 'a.ap.y': 6, 'b.bq': True, 'b.bq.y': 12,
    }
    assert c_a.to_dict() == {'ap': True, 'ap.x': 5, 'ap.y': 6}
    assert c_ap.to_dict() == {'x': 5, 'y': 6}
